/*
** Module   : WidgetRegistry.java
** Abstract :
**
** Copyright (c) 2010-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20101015          Widget registry code from Screen moved into 
**                           separate class. 
** 002 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 003 SIY 20111202          Updated imports.
** 004 VIG 20140515          Added tab order filling.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Widget registry.
 * <p>
 * This class holds all widgets (frames and their inner widgets) which were 
 * pushed from server to client. Also, this class is responsible for building
 * client side widgets from widget definition.  
 * 
 * @author 	SIY
 * @version 1.0
 */
public class WidgetRegistry
{
   /** The list of the widgets in the current screen. */
   private LinkedHashMap<Integer, Widget> widgetList = new LinkedHashMap<Integer, Widget>();

   /**
    * Constructor. 
    */
   public WidgetRegistry()
   {
   }

   /**
    * Recreates the widget based on the information can be taken from the
    * source widget snapshot. This snapshot must contain all required
    * information different from the default one which is enough to
    * reconstruct the widget of the given type with the desirable behavior.
    * 
    * @param   config
    *          Widget snapshot taken from the server.
    */
   public void reconstructWidget(ComponentConfig config)
   {
      int cid = config.getID();
      Widget comp = getComponent(cid);
   
      if (comp != null)
      {
         // existing component, just apply new config values
         Widget frame = UiUtils.locateFrame(comp);
         
         if (frame instanceof Frame)
         {
            ((Frame) frame).applyConfig(comp, config);
         }
         else
         {
            comp.<WidgetConfig>config().applyConfig(config);
         }
      }
      else
      {
         // new component
         comp = UiUtils.reconstructComponent(config);
   
         if (comp != null) 
            widgetList.put(new Integer(cid), comp);
      }
   }

   /**
    * Add pre-built frame.
    * 
    * @param   frame
    *          Frame reference.
    */
   public void addFrame(Frame frame)
   {
      if (frame == null)
         return;
   
      Widget[] list = frame.getContentPane().widgets();
      
      widgetList.put(new Integer(frame.config().getID()), frame);
   
      for (Widget comp : list)
      {
         if (UiUtils.hasConfig(comp))
         {
            widgetList.put(new Integer(comp.<WidgetConfig>config().getID()),
                           comp);
         }
      }
   }

   /**
    * Get all components defined in the screen.
    * 
    * @return  Array of component.
    */
   public Widget[] getComponents()
   {
      return widgetList.values().toArray(new Widget[widgetList.size()]);
   }

   /**
    * Remove the given frame and all contained components from the master
    * list of components.
    * 
    * @param    frame
    *           The frame to remove.
    */
   public void removeFrame(Frame frame)
   {
      if (getComponent(frame.config().getID()) != null)
      {
         int base = frame.config().getID();
         Widget comp = frame;
         
         for(int ndx = base; comp != null; ndx++)
         {
            comp = widgetList.remove(new Integer(ndx));
         }
      }
   }

   /**
    * Return <code>Component</code> instance for the given ID.
    * 
    * @param   id
    *          Component ID.
    * @return  Reference to the <code>Component</code> or <code>null</code>
    *          if no such component exists.
    */
   public Widget getComponent(int id)
   {
      if (id == WindowConfig.SESSION_WINDOW_ID)
         return Window.getDefaultWindow();
      
      return widgetList.get(new Integer(id));
   }

   /**
    * Instantiate screen from the given array of definitions.
    * 
    * @param   sd
    *          Array of data containers.
    */
   public void pushDefinition(ScreenDefinition[] sd)
   {
      for (int i = 0; i < sd.length; i++)
      {
         ScreenDefinition screenDefinition = sd[i];
   
         Frame frame = (Frame) getComponent(screenDefinition.getRootFrameID());
         
         // if frame already exists, notify it that processing begins
         if (frame != null)
            frame.preprocess();
         
         Iterator<ComponentConfig> itr = screenDefinition.getConfigsIterator();
         
         while (itr.hasNext())
         {
            reconstructWidget(itr.next());
         }
         
         if (frame == null)
            frame = (Frame) getComponent(screenDefinition.getRootFrameID());
   
         if (frame != null)
         {
            if (frame.parent() == null)
            {
               Window.getDefaultWindow().addFrame(frame, 
                                                  frame.config().getRow(),
                                                  frame.config().getColumn());
            }
            
            addWidgetsToFrame(screenDefinition.getRootFrameID(), 
                              screenDefinition);
            frame.postprocess(screenDefinition.isSuppressRedraw());
         }
      }
   }

   /**
    * Add widgets to the frame which is identified by the given frame ID.
    * 
    * @param    frameID
    *           The index of the frame to be populated.
    * @param    screen
    *           List of widget configurations.
    */
   private void addWidgetsToFrame(int frameID, ScreenDefinition screen)
   {
      ArrayList<Integer> widgetIndexes = screen.getWidgetsInFrame(frameID);
      Frame frame = (Frame) getComponent(frameID);
      Widget first = null, second = null;
   
      for (int i = 0; i < widgetIndexes.size(); i++)
      {
         Integer wid = widgetIndexes.get(i);
         int widgetIndex = wid.intValue();
         
         ComponentConfig config = screen.getConfig(widgetIndex);
         Widget nextWidget = getComponent(widgetIndex);
         
         // associate browse columns with their browse widget which is a
         // container (of sorts) itself
         if (config.getType() == ComponentConfig.BROWSE_COLUMN)
         {
            Browse browse = (Browse) 
               getComponent(((BrowseColumnConfig) config).getBrowseId());
            
            if (browse != null)
               ((BrowseColumn) nextWidget).setBrowse(browse);
   
            continue;
         }
         
         // check if this widget has already had the frame assigned as the
         // parent, if so then we MUST NOT add it to the widget list again
         // since it is already there (the Container._addComponent() method
         // sets the parent reference after unconditionally adding the 
         // component to the components vector
         if (nextWidget.parent() != null)
            continue;
   
         // make widget disabled by default
         
         if (nextWidget.isEnabled())
            nextWidget.setEnabled(false);
         
         frame.addWidget(nextWidget);
      }
      
      if (first != null && second != null)
      {
         frame.moveWidgetBeforeTabItem(first, second);
      }
      
   }

   /**
    * Get IDs of all widgets in the current screen.
    * 
    * @return  Array of all IDs.
    */
   public int[] getIDs()
   {
      Widget[] list = getComponents();
      int[] res = new int[list.length];
      
      for (int i = 0; i < res.length; i++)
         res[i] = UiUtils.getWidgetId(list[i]);
      
      return res;
   }

   /**
    * Generate an array of requested size of unused IDs.
    * 
    * @param   i
    *          Required array size.
    * @return  array of IDs.
    */
   public int[] genIDs(int i)
   {
      if (i <= 0)
         throw new RuntimeException("Invalid request size");
      
      int[] res = new int[i];
      int[] ids = getIDs();
      Arrays.sort(ids);
      
      int j = 0;
      int id = 1;
      
      while(j < res.length)
      {
         int ndx = Arrays.binarySearch(ids, id);
         
         if (ndx < 0 || ndx >= ids.length || ids[ndx] != id)
            res[j++] = id;
         
         id++;
      }
   
      return res;
   }
}
